//! a POP client for downloading messages from a popserver.
//! 
//!

#pike __REAL_VERSION__
constant __version="0.2";
constant __author="Marc Dirix <marc@electronics-design.nl";
constant __components=({"Public.pmod/Protocols.pmod/POP.pmod"});

class Protocol
{

  inherit Stdio.FILE : sock;
  string message;

  int readreturncode()
  {
    string code;

    message=sock::gets();
    if(!message) return 1;
    sscanf(message,"%3s",code);
    return code=="+OK" ? 0 : 1;
  }
  
  int command(string cmd)
  {
    sock::write(cmd+"\r\n");
    return readreturncode();
  }

  array(string) read_body_lines()
  {
    array(string) ret=({});
    string s;
    while(s = sock::gets())
    {
      if(s=="." || s==".\r") return ret;
      ret+=({s+"\n"});
    }
    error("Connection closed by server.\n");
  }

  
}

//!
class Client
{
  inherit Protocol;

  static private int cmd(string c, string|void comment)
  {
    int r = command(c);
    switch(r) {
    case 0:
      break;
    default:
      error( "POP: "+c+"\n"+(comment?"POP: "+comment+"\n":"")+
	     "POP: "+message+"\n" );
    }
    return r;
  }

//! Creates an POP mail client and connects it to the server provided.
//! If no paramaters at all are provided, the client will look up the 
//! mail host by searching for the DNS MX record.
//! 
//! @param server
//! String with the hostnaime of the popserver, or a file object acting as one.
//! @param port
//! If server is a string, than an optional port parameter may be provided. 
//! If no port parameter is provided, port 110 is assumed. 
  void create(void|string|Stdio.File server, int|void port)
  {
    if(!server)
    {
      // Lookup MX record here (Using DNS.pmod)
      object dns=master()->resolv("Protocols")["DNS"]->client();
      server=dns->get_primary_mx(gethostname());
    }

    if (objectp(server))
       assign(server);
    else
    {
       if(!port)
	  port = 110;

       if(!server || !connect(server, port))
       {
	 error("Failed to connect to POP server.\n");
       }
    }

    if(readreturncode())
      error("Connection refused by POP server.\n");

  }

//! Authenticate against the popserver using plain method.
//!
//! @param username
//! The username used in the USER command.
//! @param password
//! The corresponding password.
//! @returns
//! 0 for succes, non zero otherwise.
  int auth_plain(string username, string password)
  {
    if(cmd("user "+username))
      return 1;
    if(cmd("pass "+password)) 
         return 1;
    return 0;
  }

 //! Retrieve a list with messages and their size.
 //!
 //! @param messageid
 //! Optional parameter to only retrieve information about a given message.
 //! @returns
 //! A mapping(messageid:size)
  mapping list(int|void messageid)
  {
    mapping list = ([]);
    int id,size;
    if(!messageid)
    {
      if(cmd("LIST"))
        return ([]);
      foreach(read_body_lines(), string s)
      {
        sscanf(s,"%d %d",id,size);
        list += ([id:size]);
      }
    }
    else
    {
     if(cmd(sprintf("LIST %d",messageid)))
       return ([]);
      sscanf(message,"+OK %d %d",id,size);
        list += ([id:size]);
    }
    return list;
  }

//! Retrieve a list with messages and their unique id.
//!
//! @param messageid
//! Optional parameter to only retrieve information about a given message.
//! @returns
//! A mapping(messageid:uid) 
  mapping uidl(int|void messageid)
  {
    mapping list = ([]);
    int id;
    string uid="";
    if(!messageid)
    {
      if(cmd("UIDL"))
        return ([]);
      foreach(read_body_lines(), string s)
      {
        sscanf(s,"%d %s\r\n",id,uid);
        list += ([id:uid]);
      }
    }
    else
    {
     if(cmd(sprintf("UIDL %d",messageid)))
       return ([]);
        sscanf(message,"+OK %d %s\r",id,uid);
     list=([id:uid]);
    }
    return list;
  }

//! retrieve a message.
//! 
//! @param messageid
//! The messageid of the message to retrieve.
//! @returns
//! array(string) with messagelines.
  array retr(int messageid)
  {
    if(cmd(sprintf("RETR %d",messageid)))
      return ({});
    return read_body_lines();
  }
  
  array top(int messageid, int lines)
  {
    if(cmd(sprintf("TOP %d %d",messageid,lines)))
      return ({});
    return read_body_lines();
  }
//! Mark messages for deletion on the popserver.
//!
//! @param messagid
//! The messageid of the message to delete.
//! @returns
//! 0 for succes, non zero otherwise

  int dele(int messageid)
  {
    return cmd(sprintf("DELE %d",messageid));
  }

//! Log out of the popserver, and commit deleting messages.
//!
//! @returns
//! 0 for succes, non zero otherwise
  int quit()
  {
    return cmd("QUIT");
  }

//! No Operation. This command can be used to keep the connection allive.
//! 
//! @returns
//! 0 for succes, non zero otherwise.
  int noop()
  {
    return cmd("NOOP");
  }
};
